from __future__ import absolute_import

import datetime
from typing_extensions import Dict, List, Union, Any

from dateutil.parser import parse
from typing_extensions import Self

from mangadex.url_models import URLRequest
from .auth import Api, Auth


class Author:
    def __init__(self, auth=Union[Auth, None]) -> None:
        self.auth = auth
        self.api = Api()

        self.author_id: str = ""
        self.name: str = ""
        self.imageUrl: str = ""
        self.bio: Dict[str, str] = {}
        self.createdAt: datetime.datetime
        self.updatedAt: datetime.datetime
        self.mangas: List[str] = []

    @classmethod
    def author_from_dict(cls, resp: dict) -> Self:
        """Creates author from JSON

        Args:
            resp (dict): Raw data from JSON

        Raises:
            ValueError: Returned when the JSON doesn't contain Author Information

        Returns:
            Author: Author Information
        """
        try:
            resp = resp["data"]
        except KeyError:
            pass

        if resp["type"] != "author" or not resp:
            raise ValueError(f"The data provided is not Author is : {resp['type']}")

        author = cls()

        attributes = resp["attributes"]

        author.author_id = resp["id"]
        author.name = attributes["name"]
        author.imageUrl = attributes["imageUrl"]
        author.bio = attributes["biography"]
        author.createdAt = parse(attributes["createdAt"])
        author.updatedAt = parse(attributes["updatedAt"])
        author.mangas = [
            series["id"]
            for series in resp["relationships"]
            if series["type"] == "series"
        ]  # better keep it like this to not consume computing time

        return author

    @staticmethod
    def create_authors_list(resp: dict) -> List["Author"]:
        """Create a list of Authors from JSON

        :param resp: dict: Response from author list
        :param resp: dict:
        :returns: List[Author]: List of author

        """
        resp = resp["data"]
        authors_list = []
        for elem in resp:
            authors_list.append(Author.author_from_dict(elem))
        return authors_list

    @property
    def url(self):
        """Returns the mangadex url"""
        return f"{self.api.url}/author/{self.author_id}"

    def __eq__(self, other: Self) -> bool:
        my_vals = [self.author_id, self.name]
        other_vals = [other.author_id, other.name]
        return all((me == other for me, other in zip(my_vals, other_vals)))

    def __ne__(self, other: Self) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return (
            f"Author(id = {self.author_id}, name = {self.name}, imageUrl = {self.imageUrl},"
            f"createdAt = {self.createdAt}, updatedAt = {self.updatedAt})"
        )

    def list_author(self, **kwargs) -> List["Author"]:
        """ Get the list of authors

        Args:
            limit: Number of authors to load
            offset:
            ids[]: Array of ids
            name: Name of author(for search)

        Returns:
            List[Author]: List of Authors
        """
        if "ids" in kwargs:
            kwargs["ids[]"] = kwargs.pop("ids")

        url = f"{self.api.url}/author"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=kwargs
        )
        return list(Author.author_from_dict(author) for author in resp["data"])

    def get_author_by_id(self, author_id: str) -> "Author":
        """Gets an author by its id

        :param author_id: str`: The id of the author
        :param author_id: str:
        :returns: Author`

        """
        url = f"{self.api.url}/author/{author_id}"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return Author.author_from_dict(resp)

    def create_author(
            self, name: str, version: int, ObjReturn: bool = False
    ) -> Union["Author", None]:
        """Creates an Author

        :param name: type name: `str`. The author name
        :param version: type version: `int`. The version of the author
        :param ObjReturn: bool` Default `False`. If set to `True`, it will return the info
        :param name: str:
        :param version: int:
        :param ObjReturn: bool:  (Default value = False)
        :returns: Union[Author, None]`

        """
        url = f"{self.api.url}/author"
        params = {"name": name, "version": version}
        resp = URLRequest.request_url(
            url,
            "POST",
            timeout=self.api.timeout,
            params=params,
            headers=self.auth.get_bearer_token(),
        )
        if ObjReturn:
            return Author.author_from_dict(resp)

    def update_author(
            self,
            *,
            author_id: str,
            version: int,
            name: Union[str, None] = None,
            ObjReturn: bool = False,
    ) -> Union["Author", None]:
        """ Updates and Author

        Args:
            author_id: The author id
            version: The version of author info
            name: The name of the author
            ObjReturn: Default `False`. If set to `True`, it will return the info

        Returns:
            Union[Author, None]:  Updated Chapter
        """
        url = f"{self.api.url}/author/{author_id}"
        params: Dict[str, Any] = {"version": version}
        if name is not None:
            params["name"] = name
        resp = URLRequest.request_url(
            url, "PUT", timeout=self.api.timeout, params=params, headers=self.auth.get_bearer_token()
        )

        if ObjReturn:
            return Author.author_from_dict(resp) if ObjReturn else None

    def delete_author(self, author_id: str) -> None:
        """Deletes an author

        :param author_id: str: Author ID
        :param author_id: str:

        """
        url = f"{self.api.url}/author/{author_id}"
        URLRequest.request_url(
            url, "DELETE", headers=self.auth.get_bearer_token(), timeout=self.api.timeout
        )


class ScanlationGroup:
    def __init__(self, auth=Union[Auth, None]) -> None:
        self.auth = auth
        self.api = Api()

        self.group_id: str = ""
        self.name: List[str] = []
        self.leader = None
        self.website: str = ""
        self.discord: str = ""
        self.twitter: str = ""
        self.mangaUpdates: str = ""
        self.email: str = ""
        self.bio: str = ""
        self.focusedLanguage: str = ""
        self.official: bool = False
        self.ex_licensed: bool = False
        self.verified: bool = False
        self.inactive: bool = False
        self.publish_delay: bool = False
        self.relationships: List[Dict[str, Any]] = []

    @classmethod
    def group_from_dict(cls, resp: dict) -> Self:
        """Creates Author from JSON

        :param resp: dict: Raw data from JSON
        :param resp: dict:
        :returns: Author` Author information

        """
        try:
            resp = resp["data"]
        except KeyError:
            pass

        if resp["type"] != "scanlation_group" or not resp:
            raise ValueError(f"The data provided is not Author is : {resp['type']}",
                            )

        group = cls()

        attributes = resp["attributes"]
        group.group_id = resp["id"]
        group.name = [attributes["name"]]
        if 'altNames' in attributes:
            group.name.append([altName["name"]
                                for altName in attributes['altNames']
                                ])
        group.website = attributes["website"]
        group.discord = attributes["discord"]
        group.twitter = attributes["twitter"]
        group.mangaUpdates = attributes["mangaUpdates"]
        group.email = attributes["contactEmail"]
        group.bio = attributes["description"]
        group.focusedLanguage = attributes["focusedLanguages"]
        group.official = attributes["official"]
        if group.official == 'true':
            group.ex_licensed = attributes["exLicensed"]
        group.verified = attributes["verified"]
        group.inactive = attributes["inactive"]
        group.publish_delay = attributes["publishDelay"]
        group.relationships = [
            {relations['type']: relations['id']}
            for relations in resp["relationships"]
            if len(relations) > 1
        ]
        return group

    @staticmethod
    def create_group_list(resp) -> List["ScanlationGroup"]:
        """
        Creates a ScanlationGroup List from JSON
        """
        resp = resp["data"]
        group_list = []
        for elem in resp:
            group_list.append(ScanlationGroup().group_from_dict(elem))
        return group_list

    @property
    def url(self):
        """
        Returns the mangadex url
        """
        return f"{self.api.url}/group/{self.group_id}"

    def __eq__(self, other: Self) -> bool:
        my_vals = [self.group_id, self.name]
        other_vals = [other.group_id, other.name]
        return my_vals == other_vals

    def __ne__(self, other: Self) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return (f"ScanlationGroup(id = {self.group_id}, name = {self.name}, leader = {self.leader}, \
                createdAt = {self.createdAt}, updatedAt = {self.updatedAt})")

    def multi_group_list(self, **kwargs) -> List["ScanlationGroup"]:
        """ Get information about multiple groups

        Args:
            limit: Number of authors to load
            offset:
            ids[]: Array of ids
            name: Name of scan group(for search)

        Returns:
            List[ScanlationGroup]: List of Scanlation Groups
        """
        if "ids" in kwargs:
            kwargs["ids[]"] = kwargs.pop("ids")

        url = f"{self.api.url}/group"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=kwargs
        )
        return list(ScanlationGroup.group_from_dict(author) for author in resp["data"])

    def get_group_by_id(self, group_id: str) -> "ScanlationGroup":
        """Gets a group by its id

        :param group_id: str`: The id of the author
        
        :returns: Author`

        """
        url = f"{self.api.url}/author/{group_id}"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return ScanlationGroup.group_from_dict(resp)

    def create_group(
            self, name: str, version: int, ObjReturn: bool = False
    ) -> Union["Author", None]:
        """Creates an Author

        :param name: type name: `str`. The author name
        :param version: type version: `int`. The version of the author
        :param ObjReturn: bool` Default `False`. If set to `True`, it will return the info
        :param name: str:
        :param version: int:
        :param ObjReturn: bool:  (Default value = False)
        :returns: Union[Author, None]`

        """
        url = f"{self.api.url}/author"
        params = {"name": name, "version": version}
        resp = URLRequest.request_url(
            url,
            "POST",
            timeout=self.api.timeout,
            params=params,
            headers=self.auth.get_bearer_token(),
        )
        if ObjReturn:
            return Author.author_from_dict(resp)

    def update_group(
            self,
            *,
            author_id: str,
            version: int,
            name: Union[str, None] = None,
            ObjReturn: bool = False,
    ) -> Union["Author", None]:
        """ Updates and Author

        Args:
            author_id: The author id
            version: The version of author info
            name: The name of the author
            ObjReturn: Default `False`. If set to `True`, it will return the info

        Returns:
            Union[Author, None]:  Updated Chapter
        """
        url = f"{self.api.url}/author/{author_id}"
        params: Dict[str, Any] = {"version": version}
        if name is not None:
            params["name"] = name
        resp = URLRequest.request_url(
            url, "PUT", timeout=self.api.timeout, params=params, headers=self.auth.get_bearer_token()
        )

        if ObjReturn:
            return Author.author_from_dict(resp) if ObjReturn else None

    def delete_group(self, author_id: str) -> None:
        """Deletes an author

        :param author_id: str: Author ID
        :param author_id: str:

        """
        url = f"{self.api.url}/author/{author_id}"
        URLRequest.request_url(
            url, "DELETE", headers=self.auth.get_bearer_token(), timeout=self.api.timeout
        )


class User:
    def __init__(self, auth: Union[Auth, None]):
        self.auth = auth
        self.api = Api()

        self.id = None
        self.username = None
        self.roles = []
        self.relations = {}

    @property
    def url(self):
        """
        Returns mangadex url
        """
        return f"{self.api.url}/user/{self.id}"

    def __eq__(self, other: Self) -> bool:
        my_vals = [self.id, self.username]
        other_vals = [other.id, other.username]
        return my_vals == other_vals

    def __ne__(self, other: "User") -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return (f"User(id = {self.id}, username = {self.username}, \
                roles = {self.roles}, relations = {self.relations})")

    @classmethod
    def user_from_dict(cls, data: dict) -> "User":
        """Create a user object from raw data"""
        if "data" in data:
            data = data["data"]

        if data["type"] != "user" or not data:
            raise ValueError("The data provided is not a User JSON")

        attributes = data["attributes"]
        relationships = data["relationships"]

        user = cls(auth=None)  # No need to pass auth here
        user.id = data["id"]
        user.username = attributes["username"]
        user.roles = [' '.join([word.title() for word in role.split('_')])
                      for role in attributes["roles"]]
        user.relations = []
        if relationships:
            for relationship in relationships:
                user.relations.append({relationship["type"]: relationship["id"]})

        return user

    @staticmethod
    def create_user_list(resp: dict) -> List["User"]:
        """
        Creates a List of users from a JSON
        """
        resp = resp["data"]
        user_list = []
        for elem in resp:
            user_list.append(User.user_from_dict(elem))
        return user_list

    def get_user_list(self):
        pass

    def me(self) -> "User":
        url = f"{self.api.url}/user/me"
        resp = URLRequest.request_url(
            url, "GET",
            timeout=self.auth.timeout,
            headers=self.auth.get_bearer_token()
        )
        return User.user_from_dict(resp)

    def get_user(self, user_id: str) -> "User":
        url = f"{self.api.url}/user/{user_id}"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return User.user_from_dict(resp)


class Follows:
    def __init__(self, auth: Union[Auth, None]):
        self.auth = auth
        self.api = Api()

    def followed_groups(self, **kwargs) -> List["ScanlationGroup"]:
        """
        Get the Scanlation Groups you follow

        Parameters
        -------------
        limit `int`
        offset `int`
        translatedLanguage : `List[str]`

        Returns
        -------------
        `List[ScanlationGroup]`
        """
        if "translatedLanguage" in kwargs:
            kwargs["translatedLanguage[]"] = kwargs.pop("translatedLanguage")
        url = f"{self.api.url}/user/follows/group"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=kwargs, headers=self.auth.get_bearer_token()
        )
        return ScanlationGroup.create_group_list(resp)

    def followed_users(self, **kwargs) -> List["User"]:
        """
        Get the users you follow

        Parameters
        -------------
        limit : `int`
        offset : `int`

        Returns
        -------------
        `List[User]`
        """
        url = f"{self.api.url}/user/follows/user"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=kwargs, headers=self.auth.get_bearer_token()
        )
        return User.create_user_list(resp)

    def follow_manga(self, manga_id: Union[str, int]) -> None:
        """
        Follow a manga

        Parameters
        --------------
        manga_id : `str`. The manga id

        Raises
        -------------
        `ApiError`
        """
        url = f"{self.api.url}/manga/{manga_id}/follow"
        URLRequest.request_url(url, "POST", headers=self.auth.get_bearer_token(), timeout=self.api.timeout)

    def unfollow_manga(self, manga_id: str) -> None:
        """
        Unfollows a Manga

        Parameters
        -------------
        manga_id : `str`. The manga id

        Raises
        -----------
        `ApiError`
        """
        url = f"{self.api.url}/manga/{manga_id}/follow"
        URLRequest.request_url(url, "DELETE", headers=self.auth.get_bearer_token(), timeout=self.api.timeout)


if __name__ == '__main__':
    md = Author()
    print(md.list_author())



"""
Module for unit and integration tests
"""

import os

import pytest
from dotenv import load_dotenv

from mangadex import URLRequest, ApiError
import mangadex as md

load_dotenv()


class TestApi:
    """Class for testing API infrastructure calls"""

    api = md.Api()
    timeout = 5

    def test_ping(self):
        ping = self.api.ping()
        saved_resp = URLRequest.request_url(f"{self.api.url}/ping", "GET", self.timeout)
        assert ping == saved_resp


class TestManga:
    """
    Class for testing the public API calls
    """

    manga = md.series.Manga()
    tag = md.series.Tag()
    timeout = 5

    def test_mangaList(self):
        self.manga.get_manga_list(limit=1)[0]

    def test_SearchManga(self):
        # we're going to search for the iris zero
        resp = self.manga.get_manga_list(title="iris zero", limit=1)[0]

        url = f"{self.manga.api.url}/manga"
        raw_response = URLRequest.request_url(
            url, "GET", timeout=self.timeout, params={"limit": 1, "title": "iris zero"}
        )

        saved_resp = md.Manga.manga_from_dict(raw_response["data"][0])

        assert resp == saved_resp, "The Manga objects are not equal"

    def test_SearchMangaWithLotOfArgs(self):
        tags = self.tag.tag_list()
        wanted_tags = ["Oneshot", "Romance"]
        not_wanted_tags = ["Loli", "Incest"]
        wanted_tags_ids = []
        not_wanted_tags_ids = []
        for _, tag in enumerate(tags):
            if tag.name["en"] in wanted_tags:
                wanted_tags_ids.append(tag.tag_id)
            elif tag.name["en"] in not_wanted_tags:
                not_wanted_tags_ids.append(tag.tag_id)

        self.manga.get_manga_list(
            contentRating=["erotica", "pornographic"],
            status=["completed"],
            excludedTags=not_wanted_tags_ids,
            excludedTagsMode="AND",
            includedTags=wanted_tags_ids,
            includedTagsMode="AND",
        )

    def test_GetTranslatedManga(self):
        self.manga.get_manga_list(translatedLanguage=["en"], limit=1)[0]

    def test_GetMangaFeed(self):
        resp = self.manga.get_manga_list(title="iris zero", limit=1)[0]
        self.manga.manga_feed(resp.manga_id)

    def test_ViewMangaById(self):
        self.manga.get_manga_by_id(manga_id="88796863-04bd-49d4-ad85-d9f993e95109")

    def test_RandomManga(self):
        self.manga.random_manga()

    def test_GetMangaChaptersAndVolumes(self):
        # lets use iris zero as is in hiatus
        manga_id = "786ff721-8fd3-413d-8e50-938d8b06f917"
        resp = self.manga.get_manga_volumes_and_chapters(manga_id=manga_id)

        url = f"{self.manga.api.url}/manga/{manga_id}/aggregate"

        raw_response = URLRequest.request_url(url, "GET", timeout=self.timeout)

        saved_resp = raw_response["volumes"]

        assert resp == saved_resp, "The values are not equal"


class TestChapter:
    chapter = md.series.Chapter()
    author = md.people.Author()
    timeout = 5

    def test_GetMangaChapter(self):
        ch_id = "015979c8-ffa4-4afa-b48e-3da6d10279b0"
        resp = self.chapter.get_chapter_by_id(chapter_id=ch_id)

        url = f"{self.chapter.api.url}/chapter/{ch_id}"
        raw_response = URLRequest.request_url(
            url, "GET", timeout=self.timeout, params={"id": ch_id}
        )

        saved_resp = md.Chapter.chapter_from_dict(raw_response)

        assert resp == saved_resp, "The Chapter Objects are not equal"

    def test_FetchChapterImages(self):
        ch_id = "015979c8-ffa4-4afa-b48e-3da6d10279b0"

        resp = self.chapter.get_chapter_by_id(chapter_id=ch_id)

        resp.fetch_chapter_images()


class TestAuthor:
    author = md.people.Author()
    chapter = md.series.Chapter()
    timeout = 5

    def test_GetAuthor(self):
        author_id = "df765fdc-ea9f-45d0-9191-d95615662d49"

        resp = self.author.get_author_by_id(author_id=author_id)

        url = f"{self.author.api.url}/author/{author_id}"

        raw_response = URLRequest.request_url(
            url, "GET", timeout=self.timeout, params={"id": author_id}
        )

        saved_resp = md.Author.author_from_dict(raw_response)

        assert resp == saved_resp, "The Author Objects are not equal"

    def test_GetManyAuthor(self):
        author_ids = [
            "df765fdc-ea9f-45d0-9191-d95615662d49",
            "742bea86-c1ae-4893-bd06-805287991849",
        ]

        resp = self.author.list_author(ids=author_ids)

        url = (
            f"{self.author.api.url}/author/?ids[]={author_ids[0]}&ids[]={author_ids[1]}"
        )
        raw_response = URLRequest.request_url(url, "GET", timeout=self.timeout)

        manual_authors = [
            md.Author.author_from_dict(author) for author in raw_response["data"]
        ]
        manual_authors = dict(((author.author_id, author) for author in manual_authors))

        for author in author_ids:
            api_author = next((a for a in resp if a.author_id == author), None)
            assert api_author is not None, "Required author not returned in response"
            assert (
                api_author == manual_authors[author]
            ), "The Author Objects are not equal"


class TestTag:
    tag = md.series.Tag()
    timeout = 5

    def test_GetTags(self):
        resp = self.tag.tag_list()

        url = f"{self.tag.api.url}/manga/tag"
        raw_response = URLRequest.request_url(url, "GET", timeout=self.timeout)
        saved_reps = md.Tag.create_tag_list(raw_response)

        assert resp == saved_reps, "The test objects are not equal"


class TestScanlationGroup:
    scangroup = md.ScanlationGroup()
    api = md.Api()
    timeout = 5

    def test_GetScanlationGroup(self):
        ids = ["f5f83084-ec42-4354-96fd-1b637a89b8b3"]
        resp = self.scangroup.multi_group_list(ids=ids)  # black cat scanlations

        url = f"{self.scangroup.api.url}/group"
        raw_response = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params={"ids[]": ids}
        )

        saved_response = md.ScanlationGroup.create_group_list(raw_response)

        assert resp == saved_response


class TestCoverArt:
    manga = md.series.Manga()
    coverart = md.series.Cover()
    customlist = md.series.CustomList()

    def test_GetCoverArtList(self):
        self.coverart.get_coverart_list()

    def test_GetMangaCoverArt(self):
        random_manga = self.manga.random_manga()
        self.coverart.get_cover(random_manga.cover_id)

    def test_GetCustomList(self):
        custom_list_id = "aa0356ad-12c8-4f1a-9723-8342ade4dc6e"
        self.customlist.get_customlist(customlist_id=custom_list_id)


@pytest.mark.skipif(
    "md_username" not in os.environ
    or "md_password" not in os.environ
    or "client_id" not in os.environ
    or "client_secret" not in os.environ,
    reason="The credentials are not in env",
)
class TestAuth:
    """
    Class for testing private API calls
    """

    auth = md.auth.Auth()
    user = md.people.User(auth=auth)
    manga = md.series.Manga(auth=auth)
    follows = md.people.Follows(auth=auth)
    mangalist = md.series.MangaList(auth=auth)
    customlist = md.series.CustomList(auth=auth)

    timeout = 5

    def login(self):
        self.auth.login(
            os.environ.get("md_username"),
            os.environ.get("md_password"),
            os.environ.get("client_id"),
            os.environ.get("client_secret"),
        )

    def test_GetUser(self):
        self.login()
        user = self.user.me()

        assert user.username == os.environ.get("md_username"), "This user is invalid"

    def test_GetUserCustomLists(self):
        self.login()
        user_id = self.user.me()
        self.customlist.get_user_customlists(user_id.id)

    def test_GetMangaReadingStatus(self):
        self.login()
        manga_id = "35c33279-395d-4d9f-abec-93893c28ab29"
        self.manga.get_manga_read_markers(manga_id=manga_id)

    def test_GetAllMangaReadingStatus(self):
        self.login()
        self.manga.get_all_manga_reading_status()

    def test_GetMyMangaList(self):
        self.login()
        self.mangalist.get_my_mangalist()

    def test_FollowManga(self):
        self.login()
        manga_id = "32d76d19-8a05-4db0-9fc2-e0b0648fe9d0"  # solo leveling
        self.follows.follow_manga(manga_id=manga_id)

    def test_UnfollowManga(self):
        self.login()
        manga_id = "32d76d19-8a05-4db0-9fc2-e0b0648fe9d0"  # solo leveling
        self.follows.unfollow_manga(manga_id=manga_id)

    def test_GetMyCustomLists(self):
        self.login()
        self.customlist.get_my_customlists()


class Test_Errors:
    """
    Class for testing the errors
    """

    chapter = md.series.Chapter()
    follows = md.people.Follows(auth=None)
    timeout = 5

    def test_BadRequest(self): ...

    def test_NotFound(self):
        try:
            ch_id = (
                "015979c8-ffa4-4afa-b48e-3da6d102fdfsdfsdfsdfs"  # some random string
            )
            resp = self.chapter.get_chapter_by_id(chapter_id=ch_id)
            resp.fetch_chapter_images()
        except ApiError as e:
            assert 404 == e.code

    def test_Unauthorized(self):  # Errors out
        manga_id = "32d76d19-8a05-4db0-9fc2-e0b0648fe9d0"  # solo leveling
        try:
            self.follows.follow_manga(manga_id=manga_id)
        except AttributeError:
            pass



"""
Url handler module
"""

import json
from typing_extensions import Dict, Union, Any

import requests

from .errors import ApiError

try:
    basestring
except NameError:
    from past.builtins import basestring

try:
    from urllib.parse import urlparse, urlencode
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode


class URLRequest:
    """
    Handles the request to the server
    """

    @staticmethod
    def request_url(
        url: str,
        method: str,
        timeout,
        params: Union[Dict[str, Any], None] = None,
        headers=None,
    ) -> dict:
        """
        The handler fot GET, POST, PUT and DEL
        """
        if params is None:
            params = {}
        params = {
            k: v.decode("utf-8") if isinstance(v, bytes) else v
            for k, v in params.items()
        }

        if method == "GET":
            url = URLRequest.__build_url(url, params)
            try:
                resp = requests.get(url, headers=headers, timeout=timeout)
            except requests.RequestException as e:
                print(f"An error has occurred: {e}")
                raise
        elif method == "POST":
            try:
                resp = requests.post(url, data=params, headers=headers, timeout=timeout)
            except requests.RequestException as e:
                print(f"An error has occurred: {e}")
                raise
        elif method == "DELETE":
            try:
                resp = requests.delete(url, headers=headers, timeout=timeout)
            except requests.RequestException as e:
                print(f"An error has occurred: {e}")
                raise
        elif method == "PUT":
            try:
                resp = requests.put(
                    url, headers=headers, params=params, timeout=timeout
                )
            except requests.RequestException as e:
                print(f"An error has occurred: {e}")
                raise
        else:
            raise ValueError(f"Method {method} is invalid")

        if not resp.ok:
            raise ApiError(resp)

        content = resp.content
        data = URLRequest.__parse_data(
            content if isinstance(content, basestring) else content.decode("utf-8")
        )
        return data

    @staticmethod
    def __build_url(url: str, params: dict) -> str:
        if params and len(params) > 0:
            url = url + "?" + URLRequest.__encode_parameters(params)
        return url

    @staticmethod
    def __encode_parameters(params: dict) -> str:
        if params is None:
            return ""
        else:
            params_tuple = []
            for k, v in params.items():
                if v is None:
                    continue
                if isinstance(v, (list, tuple)):
                    for _ in v:
                        params_tuple.append((k, _))
                else:
                    params_tuple.append((k, v))
            return urlencode(params_tuple)

    @staticmethod
    def __parse_data(content):
        try:
            data = json.loads(content)
            URLRequest._check_api_error(data)
        except json.JSONDecodeError as e:
            # the ping response doen't come in JSON and its just a string return it that way and throw error in class if needed
            data = content.decode("UTF-8")
        return data

    @staticmethod
    def _check_api_error(data: dict):
        if isinstance(data, list):
            data = data[0]
        if "result" in data.keys():
            if data["result"] == "error" or "error" in data:
                raise ApiError(data["errors"])
            if isinstance(data, (list, tuple)) and len(data) > 0:
                if "error" in data:
                    raise ApiError(data["errors"])



from __future__ import absolute_import

import datetime

from dateutil.parser import parse
from typing_extensions import Dict, List, Self, Union

from mangadex.url_models import URLRequest

from .auth import Api, Auth


class Chapter:
    """Class that retrieves series chapters"""

    def __init__(self, auth: Union[Auth, None] = None) -> None:
        self.auth = auth
        self.api = Api()

        self.chapter_id: str = ""
        self.title: str = ""
        self.volume: str = ""
        self.chapter: Union[float, None] = None
        self.translatedLanguage: str = ""
        self.hash = ""
        self.data = ""
        self.manga_id: str = ""
        self.group_id: str = ""
        self.uploader: str = ""
        self.createdAt: datetime.datetime
        self.updatedAt: datetime.datetime
        self.publishAt: datetime.datetime

    # Data Processors

    @classmethod
    def chapter_from_dict(cls, resp: dict) -> "Chapter":
        """Create a Chapter from JSON

        Args:
            resp: Raw JSON data

        Returns:
            Chapter: Chapter information
        """
        chapter = cls()
        try:
            resp = resp["data"]
        except KeyError:
            pass

        if resp["type"] != "chapter" or not resp:
            raise ValueError("The data provided is not a Chapter")

        attributes = resp["attributes"]

        chapter.chapter_id = resp["id"]
        chapter.title = attributes["title"]
        chapter.volume = attributes["volume"] if not "null" else None
        chapter.chapter = (
            float(attributes["chapter"]) if attributes["chapter"] is not None else None
        )
        chapter.translatedLanguage = attributes["translatedLanguage"]
        for relations in resp["relationships"]:
            if relations["type"] == "scanlation_group":
                chapter.group_id = relations["id"]
            elif relations["type"] == "manga":
                chapter.manga_id = relations["id"]
            elif relations["type"] == "user":
                chapter.uploader = relations["id"]

        return chapter

    @staticmethod
    def __parse_chapter_list_args(params: Dict[str, str]) -> Dict[str, str]:
        if "groups" in params:
            params["groups[]"] = params.pop("groups")
        if "volume" in params:
            params["volume[]"] = params.pop("volume")
        if "translatedLanguage" in params:
            params["translatedLanguage[]"] = params.pop("translatedLanguage")

        return params

    @staticmethod
    def create_chapter_list(resp: dict) -> List["Chapter"]:
        """Creates a list of Chapters from JSON

        Args:
            resp: Raw response from chapter list

        Returns:
            List[Chapter]: List of Chapter information
        """
        resp = resp["data"]
        chap_list = []
        for elem in resp:
            chap_list.append(Chapter.chapter_from_dict(elem))
        return chap_list

    @property
    def url(self):
        """Returns the mangadex url"""
        return f"{self.api.url}/chapter/{self.chapter_id}"

    def __eq__(self, other: Self) -> bool:
        my_vals = [self.chapter_id, self.manga_id, self.chapter]
        other_vals = [other.chapter_id, other.manga_id, other.chapter]
        return my_vals == other_vals

    def __ne__(self, other: Self) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        part1 = f"Chapter(chapter_id = {self.chapter_id}, title = {self.title}, \
                        volume = {self.volume}, chapter = {self.chapter}, \
                        translatedLanguage = {self.translatedLanguage} \n"
        part2 = f"data = List[filenames], publishAt = {self.publishAt}, \
                        createdAt = {self.createdAt}, uploadedAt = {self.updatedAt}, \
                        group_id = {self.group_id}, manga_id = {self.manga_id}, uploader = {self.uploader})"
        return f"{part1}{part2}"

    def get_chapter_list(self, **kwargs) -> List["Chapter"]:
        """Get information about multiple chapters

        Args:
            limit: int: How many chapters to return
            offset: int:
            ids[]: list[str]: Chapter IDs
            groups[]: list[str]: Group UUIDs
            uploader: str: Uploader UUID
            manga: str: Manga UUID
            volume[]: str: Volume UUID
            chapter: str: Chapter UUID
            translatedLanguage[]: str: List of accepted translated language, default is any.

        Returns:
            List[Chapter]: List of Chapters
        """
        params = self.__parse_chapter_list_args(kwargs)
        url = f"{self.api.url}/chapter"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=params
        )
        return Chapter.create_chapter_list(resp)

    def get_chapter_by_id(self, chapter_id: str) -> "Chapter":
        """Get information about a single chapter

        Args:
            chapter_id: The chapter ID
            includes[]: list[str]: List of included entries. Values: manga, scanlation_group, user
        Returns:
            Chapter: Chapter info
        """
        url = f"{self.api.url}/chapter/{chapter_id}"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return Chapter.chapter_from_dict(resp)

    def get_manga_volumes_and_chapters(self, manga_id: str, **kwargs) -> Dict[str, str]:
        """Get a series volumes and chapters

        Args:
            manga_id: The series
            translatedLanguage[]: List of accepted translated language, default is any.
            groups[]: List of groups to get the releases for, default is any.

        Returns:
            Dict[str, str] List of Chapters per volume
        """
        params = None
        if "translatedLanguage" in kwargs:
            params = {"translatedLanguage[]": kwargs["translatedLanguage"]}
        url = f"{self.api.url}/manga/{manga_id}/ aggregate"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=params
        )
        return resp["volumes"]

    def fetch_chapter_images(self) -> List[str]:  # maybe make this an async function?
        """
        Get the image links for the chapter

        Returns
        -----------
        `List[str]`. A list with the links with the chapter images

        NOTE: There links are valid for 15 minutes until you need to renew the token

        Raises
        -----------
        `ApiError`
        """
        url = f"https://api.mangadex.org/at-home/server/{self.chapter_id}"
        image_server_url = URLRequest.request_url(url, "GET", timeout=5)
        self.hash = image_server_url["chapter"]["hash"]
        self.data = image_server_url["chapter"]["data"]
        image_server_url = image_server_url["baseUrl"].replace("\\", "")
        image_server_url = f"{image_server_url}/data"
        image_urls = []
        for filename in self.data:
            image_urls.append(f"{image_server_url}/{self.hash}/{filename}")

        return image_urls

    def update_chapter(
        self, chapter_id: str, body: dict, ObjReturn: bool = True
    ) -> Union["Chapter", None]:
        """Update a chapter

        Args:
            chapter_id: ID of the chapter to be updated.
            body: Body of the update.
            ObjReturn: bool: Default `True`. If set to `False`, it will not return the info

        Returns:
            Union[Chapter, None]: Updated Chapter
        """
        url = f"{self.api.url}/list/{chapter_id}"
        headers = self.auth.get_bearer_token()
        headers["Content-Type"] = "application/json"
        resp = URLRequest.request_url(
            url, "PUT", params=body, headers=headers, timeout=self.api.timeout
        )
        return self.chapter_from_dict(resp["data"]) if not ObjReturn else None

    def delete_chapter(self, chapter_id: str) -> None:
        """Delete a chapter

        Args:
            chapter_id: ID of the chapter to be deleted.

        Returns:
            dict: result
        """
        url = f"{self.api.url}/chapter/{chapter_id}"
        resp = URLRequest.request_url(
            url,
            "DELETE",
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        if resp["result"] == "error":
            raise ValueError(resp["errors"]["detail"])
        else:
            return None


class Cover:
    """Class used to get series covers."""

    def __init__(self, auth: Union[Auth, None] = None) -> None:
        self.auth = auth
        self.api = Api()

        self.cover_id: str = ""
        self.volume: str = ""
        self.fileName: str = ""
        self.description: str = ""
        self.createdAt: datetime.datetime
        self.updatedAt: datetime.datetime
        self.manga_id: str = ""
        self.locale: str = ""

    @classmethod
    def cover_from_dict(cls, data: dict) -> "Cover":
        """Get cover from json

        Args:
            data (dict): Raw JSON data

        Raises:
            ValueError: Raised when the data provided is not a Cover JSON

        Returns:
            Cover: The Coverart ids
        """
        try:
            data = data["data"]
        except (KeyError, TypeError):
            pass

        if data["type"] != "cover_art" or not data:
            raise ValueError("The data provided is not a cover")

        cover = cls()

        attributes = data["attributes"]

        cover.cover_id = data["id"]
        cover.volume = attributes["volume"]
        cover.fileName = attributes["fileName"]
        cover.locale = attributes["locale"]
        cover.description = attributes["description"]
        cover.createdAt = parse(attributes["createdAt"])
        cover.updatedAt = parse(attributes["updatedAt"])
        cover.manga_id = data["relationships"][0]["id"]

        return cover

    @staticmethod
    def create_coverart_list(resp: dict) -> List["Cover"]:
        """Creates a list of CoverArt from JSON

        Returns:
            List[Cover]: List of cover urls
        """

        resp = resp["data"]
        coverimage_list = []
        for elem in resp:
            coverimage_list.append(Cover.cover_from_dict(elem))
        return coverimage_list

    def __repr__(self) -> str:
        return f"CoverArt(id = {self.cover_id}, mangaId = {self.manga_id}, volume = {self.volume}, \
                            fileName = {self.fileName}, description = {self.description}, \
                            createdAt = {self.createdAt}, updatedAt = {self.updatedAt})"

    @staticmethod
    def __parse_coverart_params(params: Dict[str, str]) -> Dict[str, str]:
        if "manga" in params:
            params["manga[]"] = params.pop("manga")
        if "ids" in params:
            params["ids[]"] = params.pop("ids")
        if "uploaders" in params:
            params["uploaders[]"] = params.pop("uploaders")

        return params

    def fetch_cover_image(self, quality: str = "source") -> str:
        """Returns URLS of cover art

        Args:
            quality (str, optional): The quality you want to get the URL for. Defaults to "source".

        Returns:
            str: The image URL
        """

        url = f"https://uploads.mangadex.org/covers/{self.manga_id}/{self.fileName}"

        if quality == "medium":
            url = f"{url}.512.jpg"
        elif quality == "small":
            url = f"{url}.256.jpg"

        return url

    def get_coverart_list(self, **kwargs) -> List["Cover"]:
        """Gets list of CoverArt

        Returns:
            List["Cover"]: List of CoverArts
        """
        params = self.__parse_coverart_params(kwargs)
        url = f"{self.api.url}/cover"
        resp = URLRequest.request_url(
            url, "GET", params=params, timeout=self.api.timeout
        )
        return self.create_coverart_list(resp)

    def get_cover(self, cover_id: str) -> "Cover":
        """Get a cover image

        Args:
            cover_id: The cover id

        Returns:
            Cover: Cover information
        """
        url = f"{self.api.url}/cover/{cover_id}"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return self.cover_from_dict(resp)

    def upload_cover(
        self, manga_id: str, filename: str, ObjReturn: bool = False
    ) -> Union["Cover", None]:
        """Upload a Cover

        Args:
            manga_id: ID of the series you want to upload a cover
            filename: filename of the cover
            ObjReturn: Default `False`. If set to `True`, it will return the info.

        Returns:
            Union[Cover, None]: Cover info or None if ObjReturn is False
        """
        url = f"{self.api.url}/cover/{manga_id}"
        with open(filename, "rb") as f:
            file = f.read()

        resp = URLRequest.request_url(
            url,
            "POST",
            params={"file": file},
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return self.cover_from_dict(resp) if ObjReturn else None

    def edit_cover(
        self,
        cover_id: str,
        description: str,
        locale: str = "en-us",
        volume: Union[str, None] = None,
        version: Union[int, None] = None,
        ObjReturn: bool = False,
    ) -> Union[None, Self]:
        """Update a Cover Info

        Args:
            cover_id: The id of the cover
            description: The description of the cover
            locale: Language of the cover
            volume: The volume number of the cover
            version: The version number of the cover info
            ObjReturn: bool: Default `False`. If set to `True`, it will return the info

        Returns:
            Union[None, Self]: Cover information or None if ObjReturn is False
        """
        if version is None:
            raise ValueError("Version cannot be null")

        params = {"volume": volume, "locale": locale, "version": version}
        if description is not None:
            params["description"] = description

        url = f"{self.api.url}/cover/{cover_id}"
        resp = URLRequest.request_url(
            url,
            "PUT",
            params=params,
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return self.cover_from_dict(resp) if ObjReturn else None

    def delete_cover(self, cover_id: Union[str, "Cover"]):
        """Deletes a Cover

        Args:
            cover_id: ID of Cover to delete

        Returns:
            dict: result

        """
        url = f"{self.api.url}/cover/{cover_id}"
        resp = URLRequest.request_url(
            url,
            "DELETE",
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        if resp["result"] == "error":
            raise ValueError(resp["errors"]["detail"])
        else:
            return None


class Tag:
    def __init__(self) -> None:
        """Class used to get and parse tags"""

        self.api = Api()

        self.tag_id: str = ""
        self.name: Dict[str, str] = {}
        self.description: str = ""
        self.group: str = ""

    @classmethod
    def tag_from_dict(cls, resp: dict) -> "Tag":
        """Creates a Tag from a JSON

        Args:
            resp: Raw data from JSON

        Returns:
            Tag: Tag information
        """
        tag = cls()
        try:
            resp = resp["data"]
        except KeyError:
            pass

        if resp["type"] != "tag" or not resp:
            raise ValueError("The data provided is not a Tag")

        attributes = resp["attributes"]

        tag.tag_id = resp["id"]
        tag.name = attributes["name"]
        tag.description = attributes["description"]
        tag.group = attributes["group"]

        return tag

    @staticmethod
    def create_tag_list(resp) -> List["Tag"]:
        """Create a Tag list from JSON

        Args:
            resp: Response from Tag list

        Returns:
            List[Tag]: Tag list
        """
        tag_list = []
        try:
            resp = resp["data"]
        except (TypeError, KeyError):
            pass

        for tag in resp:
            tag_list.append(Tag.tag_from_dict(tag))
        return tag_list

    def __eq__(self, other: Self) -> bool:
        my_vals = [self.tag_id, self.name]
        other_vals = [other.tag_id, other.name]
        return all((me == other for me, other in zip(my_vals, other_vals)))

    def __ne__(self, other: Self) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"Tag(tag_id = {self.tag_id}, name = {self.name})"

    def tag_list(self) -> List["Tag"]:
        """Get the list of available tags

        Returns:
            List[Tag]: Tag list
        """
        url = f"{self.api.url}/manga/tag"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return Tag.create_tag_list(resp)


class Manga:
    def __init__(self, auth: Union[Auth, None] = None):
        self.auth = auth
        self.api = Api()

        self.manga_id: str = ""
        self.title: Dict[str, str] = {}
        self.altTitles: Dict[str, str] = {}
        self.description: Dict[str, str] = {}
        self.isLocked: bool = False
        self.links: Dict[str, str] = {}
        self.originalLanguage: str = ""
        self.lastVolume: str = ""
        self.lastChapter: str = ""
        self.publicationDemographic: str = ""
        self.status: str = ""
        self.year: int = 0
        self.contentRating: str = ""
        self.tags: List[Tag] = []
        self.version = 1
        self.createdAt: datetime.datetime
        self.updatedAt: datetime.datetime
        self.author_id: List[str] = []
        self.artist_id: List[str] = []
        self.cover_id: str = ""

    @classmethod
    def manga_from_dict(cls, data: dict):
        """
        Creates a Manga Object from a JSON
        """
        try:
            data = data["data"]
        except (TypeError, KeyError):
            pass

        if data["type"] != "manga" or not data:
            raise ValueError("The data provided is not a Manga")

        attributes = data["attributes"]

        manga = cls()

        manga.manga_id = data["id"]
        manga.title = attributes["title"]
        manga.altTitles = attributes["altTitles"]
        manga.description = attributes["description"]
        try:
            manga.isLocked = attributes["isLocked"]
        except KeyError:
            pass

        manga.links = attributes["links"]
        manga.originalLanguage = attributes["originalLanguage"]
        manga.lastVolume = attributes["lastVolume"]
        manga.lastChapter = attributes["lastChapter"]
        manga.publicationDemographic = attributes["publicationDemographic"]
        manga.status = attributes["status"]
        manga.year = attributes["year"]
        manga.contentRating = attributes["contentRating"]
        manga.tags = Tag.create_tag_list(attributes["tags"])
        manga.createdAt = parse(attributes["createdAt"])
        manga.updatedAt = parse(attributes["updatedAt"])

        for elem in data["relationships"]:
            if elem["type"] == "author":
                manga.author_id.append(elem["id"])
            elif elem["type"] == "artist":
                manga.artist_id.append(elem["id"])
            elif elem["type"] == "cover_art":
                manga.cover_id = elem["id"]

        return manga

    @staticmethod
    def __parse_manga_params(params: dict) -> dict:
        if "authors" in params:
            temp = params.pop("authors")
            params["authors[]"] = temp
        if "artist" in params:
            temp = params.pop("artist")
            params["artist[]"] = temp
        if "excludedTags" in params:
            temp = params.pop("excludedTags")
            params["excludedTags[]"] = temp
        if "originalLanguage" in params:
            temp = params.pop("originalLanguage")
            params["originalLanguage[]"] = temp
        if "includedTags" in params:
            temp = params.pop("includedTags")
            params["includedTags[]"] = temp
        if "publicationDemographic" in params:
            temp = params.pop("publicationDemographic")
            params["publicationDemographic[]"] = temp
        if "ids" in params:
            temp = params.pop("ids")
            params["ids[]"] = temp
        if "altTitles" in params:
            temp = params.pop("altTitles")
            params["altTitles[]"] = temp
        if "description" in params:
            temp = params.pop("description")
            params["description[]"] = temp
        if "authors" in params:
            temp = params.pop("authors")
            params["authors[]"] = temp
        if "artists" in params:
            temp = params.pop("artists")
            params["artists[]"] = temp
        if "translatedLanguage" in params:
            temp = params.pop("translatedLanguage")
            params["availableTranslatedLanguage[]"] = temp
        if "status" in params:
            params["status[]"] = params.pop("status")
        if "contentRating" in params:
            params["contentRating[]"] = params.pop("contentRating")
        return params

    @staticmethod
    def create_manga_list(resp) -> List["Manga"]:
        """
        Creates a manga list from a JSON
        """
        resp = resp["data"]
        manga_list = []
        for elem in resp:
            manga_list.append(Manga.manga_from_dict(elem))
        return manga_list

    @property
    def url(self):
        """
        Return the mangadex url
        """
        return f"{self.api.url}/title/{self.manga_id}"

    def __eq__(self, other: Self) -> bool:
        my_vals = [self.manga_id, self.title, self.createdAt, self.author_id]
        other_vals = [other.manga_id, other.title, other.createdAt, other.author_id]
        return my_vals == other_vals

    def __ne__(self, other: Self) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        temp1 = f"Manga(manga_id = {self.manga_id}, title = {self.title}, \
                altTitles = {self.altTitles}, description = {self.description}, \
                isLocked = {self.isLocked}, links = {self.links}, \
                originalLanguage = {self.originalLanguage} \n"
        temp2 = f"lastVolume = {self.lastVolume}, lastChapter = {self.lastChapter}, \
                publicationDemographic = {self.publicationDemographic}, status = {self.status}, \
                year = {self.year}, contentRating = {self.contentRating} \n"
        temp3 = f"createdAt = {self.createdAt}, uploadedAt = {self.updatedAt}), \
                author_id = {self.author_id}, artist_id = {self.artist_id}, \
                cover_id = {self.cover_id}"
        return f"{temp1}{temp2}{temp3}"

    def get_manga_list(self, **kwargs) -> List["Manga"]:
        """
        Search a List of Manga

        Parameters
        -------------
        This parameters may be used by other methods
        ### QueryParams:

        limit : `int`
        offset : `int`
        title : `str`
        authors : `List[str]`
        artist : `List[str]`
        year : `int`
        includedTags : `List[Tag.id]`
        includedTagsMode: `str`. Default `"AND"`. Enum: `"AND"` `"OR"`
        excludedTags : `List[Tag.id]`
        excludedTagsMode : `str`. Default `"AND"`, Enum : `"AND"`, `"OR"`
        status : `List[str]`. Items Enum : `"ongoing"`, `"completed"`, `"hiatus"`, `"cancelled"`
        originalLanguage : `List[str]`
        publicationDemographic : `List[str]`. Items Enum: `"shounen"` `"shoujo"` `"josei"` `"seinen"` `"none"`
        manga_ids :  `List[str]`. Limited to 100 per call
        contentRating : `List[str]`. Items Enum : `"safe"` `"suggestive"` `"erotica"` `"pornographic"`
        createdAtSince : `str`. Datetime String with the following format YYYY-MM-DDTHH:MM:SS
        updatedAtSince : `str`. Datetime String with the following format YYYY-MM-DDTHH:MM:SS

        Returns
        -------------
        `List[Manga]`. A list of Manga objects

        Raises
        -------------
        `ApiError` `MangaError`
        """
        params = kwargs
        params = self.__parse_manga_params(params)
        url = f"{self.api.url}/manga"
        resp = URLRequest.request_url(
            url, "GET", params=params, timeout=self.api.timeout
        )
        return Manga.create_manga_list(resp)

    def manga_feed(self, manga_id: str, **kwargs) -> List[Chapter]:
        """
        Get the manga feed

        Parameters
        ------------
        manga_id `str`, Required. The manga id

        ### QueryParams:

        limit : `int`
        offset : `int`
        translatedLanguage : `List[str]`. The translated languages to query
        createdAtSince : `str`. Datetime String with the following format YYYY-MM-DDTHH:MM:SS
        updatedAtSince : `str`. Datetime String with the following format YYYY-MM-DDTHH:MM:SS

        Returns
        -------------
        `List[Chapter]` A list of Chapter Objects

        Raises
        -------------
        `ApiError` `ChapterError`

        Args:
            manga_id:
            manga_id:
        """
        kwargs = self.__parse_manga_params(kwargs)
        url = f"{self.api.url}/manga/{manga_id}/feed"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=kwargs
        )
        return Chapter.create_chapter_list(resp)

    def get_manga_by_id(self, manga_id: str) -> "Manga":
        """
        Get a Manga by its id

        Parameters
        ------------
        manga_id: `str`. The manga id

        Returns
        -------------
        `Manga`. A Manga object

        Raises
        ------------
        `ApiError` `MangaError`
        """
        url = f"{self.api.url}/manga/{manga_id}"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return Manga.manga_from_dict(resp)

    def random_manga(self) -> "Manga":
        """
        Get a random Manga

        Returns
        ----------
        `Manga`. A Manga object

        Raises
        ----------
        `ApiError` `MangaError`
        """
        url = f"{self.api.url}/manga/random"
        resp = URLRequest.request_url(url, "GET", timeout=self.api.timeout)
        return Manga.manga_from_dict(resp)

    def create_manga(self, title: str, **kwargs) -> "Manga":
        """
        Creates a manga

        Parameters
        -----------
        title : `str`. The manga title

        ### Optional Parameters
        altTitles : `List[Dict[str,str]]`. The alt titles
        description : `Dict[str,str]`. The alt titles in different languages
        authors : `List[str]`. The list of author id's
        artists : `List[str]`. The list of artist id's
        links : `Dict[str,str]`. The links in different sites (al, ap, bw, mu, etc.).
        Please refer to the [documentation](https://api.mangadex.org/docs.html#section/Static-data/Manga-links-data)
        originalLanguage : `str`. The original Language
        lastVolume : `str`. The last volume
        lastChapter : `str`. The last chapter
        publicationDemographic : `str`.
        status : `str`.
        year : `int`.
        contentRating : `str`.
        modNotes : `str`

        Returns
        ------------
        `Manga`. A manga object if `ObjReturn` is set to `True`
        """
        params = self.__parse_manga_params(kwargs)
        url = f"{self.api.url}/manga"
        params["title"] = title
        resp = URLRequest.request_url(
            url,
            "POST",
            params=params,
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return Manga.manga_from_dict(resp)

    def get_manga_volumes_and_chapters(self, manga_id: str, **kwargs) -> Dict[str, str]:
        """
        Get a manga volumes and chapters

        Parameters
        ------------
        manga_id : `str`. The manga id
        translatedLanguage : `List[str]`

        Returns
        ------------
        `Dict[str, str]`. A dictionary with the volumes and the chapter id's
        """
        params = None
        if "translatedLanguage" in kwargs:
            params = {"translatedLanguage[]": kwargs["translatedLanguage"]}
        url = f"{self.api.url}/manga/{manga_id}/aggregate"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=params
        )
        return resp["volumes"]

    def update_manga(
        self, manga_id: str, ObjReturn: bool = False, **kwargs
    ) -> Union["Manga", None]:
        """
        Updates a manga parameters

        Parameters
        -------------
        ### Required parameters
        manga_id : `str`. The manga id
        version : `int`
        ObjReturn : `bool`. `True` if you want a Manga Object return

        ### Optional Parameters
        title : `Dict[str,str]`. The manga title
        altTitles : `List[Dict[str,str]]`. The alt titles
        description : `Dict[str,str]`. The alt titles in different languages
        authors : `List[str]`. The list of author id's
        artists : `List[str]`. The list of artist id's
        links : `Dict[str,str]`. The links in different sites (al, ap, bw, mu, etc.).
            Please refer to the [documentation](https://api.mangadex.org/docs.html#section/Static-data/Manga-links-data)
        originalLanguage : `str`. The original Language
        lastVolume : `str`. The last volume
        lastChapter : `str`. The last chapter
        publicationDemographic : `str`.
        status : `str`.
        year : `int`.
        contentRating : `str`.
        modNotes : `str`

        Returns
        ------------
        `Manga`. A manga object if `ObjReturn` is set to `True`
        """
        kwargs = self.__parse_manga_params(kwargs)
        url = f"{self.api.url}/manga/{manga_id}"
        resp = URLRequest.request_url(
            url,
            "PUT",
            params=kwargs,
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        if ObjReturn:
            return Manga.manga_from_dict(resp)

    def delete_manga(self, manga_id: str) -> None:
        """
        Deletes a manga

        Parameters
        ------------
        manga_id : `str`. The manga id
        """
        url = f"{self.api.url}/manga{manga_id}"

        URLRequest.request_url(
            url,
            "DELETE",
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )

    def get_manga_read_markers(self, manga_id: str) -> List[Chapter]:
        # this needs a performance update
        """
        A list of Chapter ids That are marked from the given manga id

        Parameters
        ------------
        manga_id : `str`. The Manga id

        Returns
        -------------
        `List[Chapters]`. A list of chapters that are marked as read
        """
        url = f"{self.api.url}/manga/{manga_id}/read"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, headers=self.auth.get_bearer_token()
        )
        chap_ids = resp["data"]
        chapter = Chapter()
        return [chapter.get_chapter_by_id(chap_id) for chap_id in chap_ids]

    def get_manga_reading_status(self, manga_id: Union[str, int]) -> str:
        """
        Get a manga reading status given its id

        Parameters
        ------------
        manga_id : `str`. The manga id

        Returns
        ------------
        `str` The manga reading status
        """
        url = f"{self.api.url}/manga/{manga_id}/status"
        resp = URLRequest.request_url(
            url, "GET", headers=self.auth.get_bearer_token(), timeout=self.api.timeout
        )
        return resp["status"]

    def get_all_manga_reading_status(
        self, status: Union[str, None] = None
    ) -> Dict[str, str]:
        """
        Get all Manga followed by the user reading status

        Parameters
        ------------
        status : `str`. Optional.
            Values : `"reading"` `"on_hold"` `"plan_to_read"` `"dropped"` `"re_reading"` `"completed"`

        Returns
        -----------
        `Dict[str,str]` A dictionary with the Manga id and its status
        """
        url = f"{self.api.url}/manga/status"
        resp = URLRequest.request_url(
            url,
            "GET",
            params={"status": status},
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return resp["statuses"]

    def update_manga_reading_status(self, manga_id: str, status: str) -> None:
        """
        Update the reading status of a manga

        Parameters
        -------------
        manga_id : `str`. The manga id.\n
        status : `str`. Values : `"reading"` `"on_hold"` `"plan_to_read"` `"dropped"` `"re_reading"` `"completed"`

        Raises
        -------------
        `ApiError`
        """
        url = f"{self.api.url}/manga/{manga_id}/status"
        URLRequest.request_url(
            url,
            "POST",
            params={"status": status},
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )


class MangaList(Manga):
    def __init__(self, auth=Auth):
        super().__init__(auth=auth)

    def get_my_mangalist(self, **kwargs) -> List["Manga"]:
        url = f"{self.api.url}/user/follows/manga"
        resp = URLRequest.request_url(
            url,
            "GET",
            timeout=self.api.timeout,
            params=kwargs,
            headers=self.auth.get_bearer_token(),
        )
        return self.create_manga_list(resp)


class CustomList:
    def __init__(self, auth: Union[Auth, None] = None):
        self.auth = auth
        self.api = Api()
        self.list_id: str = ""
        self.name: str = ""
        self.visibility: str = ""
        self.owner: str = ""
        self.mangas: List[str] = []

    @classmethod
    def list_from_dict(cls, data: dict) -> Self:
        """
        Creates a CustomList from a JSON
        """
        if data["type"] != "custom_list" or not data:
            raise ValueError(data, "The data provided is not a CustomList")

        custom_list = cls()

        attributes = data["attributes"]
        relationships = data["relationships"]

        custom_list.list_id = data["id"]
        custom_list.name = attributes["name"]
        custom_list.visibility = attributes["visibility"]
        for elem in relationships:
            if elem["type"] == "user":
                custom_list.owner = elem["id"]
            elif elem["type"] == "manga":
                custom_list.mangas.append(elem["id"])

        return custom_list

    @staticmethod
    def create_customlist_list(resp) -> List["CustomList"]:
        """
        Creates a list of CustomList from a JSON
        """
        resp = resp["data"]
        custom_lists = []
        for elem in resp:
            custom_lists.append(CustomList.list_from_dict(elem))
        return custom_lists

    def __repr__(self) -> str:
        return f"CustomList(id = {self.list_id}, name = {self.name},\
                visibility = {self.visibility}, owner = {self.owner}, Manga = List[Manga])"

    def get_my_customlists(self, **kwargs) -> List["CustomList"]:
        """
        Get my custom lists

        Parameters
        ------------
        ### QueryParams:
        limit : `int`. The limit of custom lists to return
        offset : `int`. The amount of offset

        Returns
        ----------
        `List[CustomList]`
        """
        url = f"{self.api.url}/user/list"
        resp = URLRequest.request_url(
            url,
            "GET",
            params=kwargs,
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return self.create_customlist_list(resp)

    def get_user_customlists(self, user_id: str, **kwargs) -> List["CustomList"]:
        """
        Get a user's custom list. This will list only public custom lists

        Parameters
        ------------
        user_id : `str`. the User id

        ### QueryParams:
        limit : `int`. The limit of custom lists to return
        offset : `int`. The amount of offset

        Returns
        ----------
        `List[CustomList]`
        """
        url = f"{self.api.url}/user/{user_id}/list"
        resp = URLRequest.request_url(
            url,
            "GET",
            params=kwargs,
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return self.create_customlist_list(resp)

    def add_manga_to_customlist(self, manga_id: str, list_id: str) -> None:
        """
        Adds a manga to a custom list

        Parameters
        --------------
        manga_id : `str`. The manga id.
        list_id : `str`. The list id.
        """
        url = f"{self.api.url}/{manga_id}/list{list_id}"
        URLRequest.request_url(
            url, "POST", headers=self.auth.get_bearer_token(), timeout=self.api.timeout
        )

    def remove_manga_from_customlist(self, manga_id: str, list_id: str) -> None:
        """
        Removes a manga from a custom list

        Parameters
        ------------
        manga_id : `str`. The manga id
        list_id : `str`. The list id
        """
        url = f"{self.api.url}/manga/{manga_id}/list/{list_id}"
        URLRequest.request_url(
            url,
            "DELETE",
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )

    def create_customlist(
        self,
        name: str,
        visibility: str = "public",
        manga: Union[List[str], None] = None,
        version: int = 1,
    ) -> None:
        """
        Creates a custom list

        Parameters
        -------------
        ### QueryParams:

        name : `str`. The custom list name
        visibility : `str. The visibility of the custom list
        manga : `List[str]`. List of manga ids
        """
        url = f"{self.api.url}/list"
        params = {
            "name": name,
            "version": version,
            "visibility": visibility,
            "manga[]": manga,
        }
        URLRequest.request_url(url, "POST", params=params, timeout=self.api.timeout)

    def get_customlist(self, customlist_id: str, **kwargs) -> "CustomList":
        """
        Get a custom list by its id

        Parameters
        ------------
        customlist_id : `str`. The id of the custom list
        limit : int
        offset : int
        translatedLanguage : List[str]
        createdAtSince : `str`. Datetime String with the following format YYYY-MM-DDTHH:MM:SS
        updatedAtSince : `str`. Datetime String with the following format YYYY-MM-DDTHH:MM:SS
        publishAtSince : `str`. Datetime String with the following format YYYY-MM-DDTHH:MM:SS
        Returns
        ------------
        `CustomList`
        """
        url = f"{self.api.url}/list/{customlist_id}"
        resp = URLRequest.request_url(
            url, "GET", timeout=self.api.timeout, params=kwargs
        )
        return CustomList.list_from_dict(resp["data"])

    def update_customlist(self, customlist_id: str, **kwargs) -> "CustomList":
        """
        Update a custom list

        Parameters
        ------------
        customlist_id : `str`. The custom list id

        ### QueryParams:
        name : `str`. The custom list name
        visibility : `str`. Values : `"public"` `"private"`

        Returns
        -----------
        `CustomList`
        """
        url = f"{self.api.url}/list/{customlist_id}"
        resp = URLRequest.request_url(
            url,
            "PUT",
            params=kwargs,
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return CustomList.list_from_dict(resp["data"])

    def delete_customlist(self, customlist_id: str) -> None:
        """
        Deletes a Custom List

        Parameters
        ------------
        customlist_id : `str`. The custom list id
        """
        url = f"{self.api.url}/list{customlist_id}"
        URLRequest.request_url(
            url,
            "DELETE",
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )



"""
Python module for interacting with the mangadex API
"""
from .errors import (
    ApiError
)
from .url_models import URLRequest

from .auth import (
    Api,
    ApiClient,
    Auth
)
from .people import (
    Author,
    ScanlationGroup,
    User,
    Follows
)
from .series import (
    Chapter,
    Cover,
    Tag,
    Manga,
    MangaList,
    CustomList,
)
__author__ = "Eduardo Ceja"
__version__ = "2.7"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2021 Eduardo Ceja"



from __future__ import absolute_import

from typing_extensions import Self, Union, Dict, List

from mangadex.errors import ApiError
from mangadex.url_models import URLRequest


# TODO: Fix the in-code documentation
class Api:
    def __init__(self):
        self.url = "https://api.mangadex.org"
        self.timeout = 5

    def ping(self):
        url = f"{self.url}/ping"
        pong = URLRequest.request_url(url, "GET", timeout=self.timeout)
        if pong != "pong":
            raise ApiError(
                {"status": "503", "reason": "Infrastructure Unavailable"},
                "MangaDex Infrastructure is down",
            )
        # if we raise we dont return
        return pong


class Auth:
    def __init__(self):
        """Authentication class"""
        self.auth_url = "https://auth.mangadex.org"
        self.timeout = 5  # Default timeout
        self.bearer = None
        self.refresh_token = None
        self.client_id = None
        self.client_secret = None

    def set_bearer_token(self, bearer_token):
        self.bearer = bearer_token

    def get_bearer_token(self):
        return self.bearer

    def __auth_handler(self, http_form: dict, headers: dict) -> None:
        """Handle OAuth2 Requests"""
        url = f"{self.auth_url}/realms/mangadex/protocol/openid-connect/token"
        auth_response = URLRequest.request_url(
            url, "POST", params=http_form, timeout=self.timeout, headers=headers
        )

        self.client_id = http_form["client_id"]
        self.client_secret = http_form["client_secret"]
        access_token = auth_response["access_token"]
        refresh_token = auth_response["refresh_token"]
        self.set_bearer_token({"Authorization": f"Bearer {access_token}"})
        self.refresh_token = refresh_token

    def login(
        self, username: str, password: str, client_id: str, client_secret: str
    ) -> None:
        """Logs into MangaDex

        Args:
            username (str): The user's username
            password (str): The user's password
            client_id (str): The user's personal client ID
            client_secret (str): The user's personal client secret
        """
        headers = {"Content-type": "application/x-www-form-urlencoded"}
        self.__auth_handler(
            {
                "grant_type": "password",
                "username": username,
                "password": password,
                "client_id": client_id,
                "client_secret": client_secret,
            },
            headers,
        )

    def refresh_login(self) -> None:
        """Reauthenticate using refresh token"""
        headers = {"Content-type": "application/x-www-form-urlencoded"}
        self.__auth_handler(
            {
                "grant_type": "refresh_token",
                "refresh_token": self.refresh_token,
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            },
            headers,
        )


class ApiClient(Auth):
    def __init__(self, auth: Union[Auth, None]):
        super().__init__()
        self.api = api
        self.auth = auth

        self.name = ""
        self.description = ""
        self.profile = ""
        self.client_id = ""
        self.state = ""
        self.isActive = ""
        self.createdAt = ""
        self.updatedAt = ""
        self.relations = []

    @classmethod
    def client_from_dict(cls, data: dict):
        """Get client from json

        Args:
            data (dict): Raw JSON data

        Raises:
            ValueError: Raised when the data provided is not a Client JSON

        Returns:
            Cover: The Coverart ids
        """
        try:
            data = data["data"]
        except (KeyError, TypeError):
            pass

        if data["type"] != "api_client" or not data:
            raise ValueError("The data provided is not a api_client")

        client = cls(auth=None)

        attributes = data["attributes"]

        client.name = attributes["name"]
        client.description = attributes["description"]
        client.profile = attributes["profile"]
        client.client_id = attributes["externalClientId"]
        client.state = attributes["state"]
        client.isActive = attributes["isActive"]
        client.createdAt = attributes["createdAt"]
        client.updatedAt = attributes["updatedAt"]
        for relation in attributes["relationships"]:
            client.relations.append({"type": relation["type"], "id": relation["id"]})

        return client

    def __eq__(self, other: Self) -> bool:
        my_vals = [self.client_id, self.name, self.createdAt]
        other_vals = [other.client_id, other.name, other.createdAt]
        return my_vals == other_vals

    def __ne__(self, other: Self) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        part1 = f"ApiClient(clientId = {self.client_id}, name = {self.name}, \
                        description = {self.description}, profile = {self.profile}, \
                        state = {self.state}, isActive = {self.isActive} \n"
        part2 = f"createdAt = {self.createdAt}, uploadedAt = {self.updatedAt} \
                        relations = {self.relations})"
        return f"{part1}{part2}"

    @staticmethod
    def __parse_client_params(params: Dict[str, str]) -> Dict[str, str]:
        if "includes" in params:
            params["includes[]"] = params.pop("includes")
        return params

    @staticmethod
    def create_client_list(resp: dict):
        resp = resp["data"]
        client = []
        for elem in resp:
            client.append(ApiClient.client_from_dict(elem))
        return client

    def get_api_clients(self, **kwargs) -> List["ApiClient"]:
        """Get users ApiClient

        Args:
            **kwargs: The API client arguments

        Returns:
            ApiClient:
        """
        params = self.__parse_client_params(kwargs)
        url = f"{self.api.url}/client"
        resp = URLRequest.request_url(
            url,
            "GET",
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
            params=params,
        )
        return ApiClient.create_client_list(resp)

    def get_api_client_by_id(self, client_id: str):
        url = f"{self.api.url}/client/{client_id}"
        resp = URLRequest.request_url(
            url, "GET", headers=self.auth.get_bearer_token(), timeout=self.timeout
        )
        return ApiClient.client_from_dict(resp)

    def create_api_client(
        self, name: str, description: str, ObjReturn: bool = True
    ) -> Union["ApiClient", None]:
        url = f"{self.api.url}/client"
        params = {
            "name": name,
            "description": description,
            "profile": "personal",
            "version": 1,
        }
        resp = URLRequest.request_url(
            url,
            "POST",
            timeout=self.api.timeout,
            headers=self.auth.get_bearer_token(),
            params=params,
        )
        return ApiClient.client_from_dict(resp) if ObjReturn else None

    def edit_api_client(self, client_id: str, description: str, ObjReturn: bool = True):
        url = f"{self.api.url}/client/{client_id}"
        params = {"description": description}
        resp = URLRequest.request_url(
            url,
            "GET",
            params=params,
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        return ApiClient.client_from_dict(resp) if ObjReturn else None

    def delete_api_client(self, client_id: str):
        url = f"{self.api.url}/client/{client_id}"
        resp = URLRequest.request_url(
            url,
            "DELETE",
            headers=self.auth.get_bearer_token(),
            timeout=self.api.timeout,
        )
        if resp["result"] != "ok":
            raise Exception(resp["errors"][0]["detail"])

    def get_api_secret(self, client_id: str):
        url = f"{self.api.url}/client/{client_id}/secret"
        resp = URLRequest.request_url(
            url, "GET", headers=self.auth.get_bearer_token(), timeout=self.timeout
        )
        if resp["result"] != "ok":
            raise Exception(resp["errors"][0]["detail"])
        return ApiClient.client_from_dict(resp)

    def regen_api_secret(self, client_id: str):
        url = f"{self.api.url}/client/{client_id}/secret"
        resp = URLRequest.request_url(
            url, "POST", headers=self.auth.get_bearer_token(), timeout=self.timeout
        )
        if resp["result"] != "ok":
            raise Exception(resp["errors"][0]["detail"])
        return ApiClient.client_from_dict(resp)


if __name__ == "__main__":
    from .people import User
    from dotenv import load_dotenv

    load_dotenv("mangadex/api/.env")
    api = Api()
    # auth = Auth()
    # auth.login(os.environ['md_username'], os.environ['md_password'],
    #            os.environ['client_id'], os.environ['client_secret'])
    user = User(auth=None)
    print(user.get_user("fca42d9d-b00c-4b52-8966-c3e1061fca4e"))



"""
Python setup for the library
"""
from setuptools import setup, find_packages

VERSION = "2.7"
DESCRIPTION = "Python wrapper for the mangadex API"
with open("README.md", "r", encoding="utf8") as f:
    LONG_DESCRIPTION = f.read()

# setting up
setup(
    name="mangadex",
    version=VERSION,
    author="Eduardo Ceja",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        "requests",
        "future",
        "python-dateutil",
        "pytest",
        "typing-extensions",
    ],
    source="https://github.com/EMACC99/mangadex",
    download_url="https://github.com/EMACC99/mangadex/releases",
    documentation="https://github.com/EMACC99/mangadex/wiki",
    license="MIT",
    keywords=["python", "mangadex"],
    clasifiers=[
        "License :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet",
        "Topic :: Library",
        "Topic :: Wrapper",
    ],
    zip_safe=False,
    python_requires=">=3.6",
)



"""
Module for error class declaration
"""
from typing_extensions import Union

from requests import Response


class ApiError(Exception):
    def __init__(
            self, resp: Union[Response, dict], message="The api responded with the error"
    ) -> None:
        self.resp = resp
        self.details = ""
        if isinstance(self.resp, Response):
            self.code = self.resp.status_code
            self.details = self.resp.reason

        else:
            self.code = self.resp["status"]
            self.details = self.resp["reason"]

        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"{self.message}: {self.code} \n {self.details}"


class BaseError(Exception):
    def __init__(self, data: dict, message: str) -> None:
        self.data = data
        self.message = message
        super(BaseError, self).__init__(self.message)


class ApiClientError(BaseError):
    def __init__(self, data: dict, message: str) -> None:
        super(ApiClientError, self).__init__(data, message=message)
        self.data = data
        self.message = message
